import os

from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import _pytest
import pluggy

from _pytest._code.code import ExceptionInfo
from cleo.formatters.formatter import Formatter
from cleo.io.outputs.output import Output
from cleo.io.outputs.section_output import SectionOutput

import treat

from treat.ui.exception_trace import ExceptionTrace

from .state import State
from .test_report import TestReport
from .test_result import TestResult


class Style:

    _TYPES = [TestResult.PASS, TestResult.FAIL, TestResult.SKIPPED]

    def __init__(self, output: Output) -> None:
        self._output = output
        self._footer: SectionOutput = self._output.section()
        self._ignore_files_in = r"^({}|{}|{}|{})".format(
            os.path.dirname(_pytest.__file__),
            os.path.dirname(pluggy.__file__),
            treat.__file__,
            os.path.join(os.path.dirname(treat.__file__), "mock"),
        )
        self._inline_errors = False
        self._errors: Dict[Tuple[str], List[TestResult]] = {}
        self._errors_count = 0

    @property
    def footer(self) -> SectionOutput:
        return self._footer

    def inline_errors(self, inline_errors: bool = True) -> "Style":
        self._inline_errors = inline_errors

        return self

    def write_current_recap(self, state: State) -> None:
        if not state.group_tests_count:
            return

        self._footer.clear()

        self._output.write_line(
            self._title_line(
                "black", state.group_title_color, state.group_title, state.group
            )
        )

        for test in state.tests:
            description = test.description

            if test.parameters:
                description += " <fg=blue><b>#</b>{}</blue>".format(
                    state.parameter_set(test)
                )

            self._output.write_line(
                self._test_line(
                    test.color,
                    test.icon,
                    description,
                    warning=test.skip_reason if test.is_skipped() else None,
                )
            )
            if self._output.is_very_verbose():
                self._output.write_line("    " + self._identifier_line(test.identifier))

                self._write_parameters(test)

            if test.error:
                self._errors.setdefault(state.group, [])
                self._errors[state.group].append(test)
                self._errors_count += 1

                if self._inline_errors:
                    self._write_error(test)

    def update_footer(self, state: State, report: Optional[TestReport] = None) -> None:
        self._footer.overwrite(self._footer_content(state, report))

    def write_recap(self, state: State, elapsed: float) -> None:
        self._footer.clear()

        if self._errors and not self._inline_errors:
            self.write_errors_summary()

        self._output.write_line(self._footer_content(state))
        self._output.write_line(
            f"  <fg=default;options=bold>Time:   </><fg=default>{elapsed:0.2f}s</>"
        )

    def write_errors_summary(self) -> None:
        i = 1
        padding = len(str(self._errors_count))
        for group, errors in self._errors.items():
            for test in errors:
                self._output.write_line("")
                if len(self._errors) > 1:
                    self._output.write_line("")
                    line = (
                        f"<fg=red;options=bold>Failing</> test "
                        f"<fg=default;options=dark>(</><fg=default;options=bold>{i:{padding}}<fg=default;options=dark>/</>{self._errors_count}</><fg=default;options=dark>)</>"
                    )
                    self._output.write_line("  <fg=default;options=bold># " + line)
                    self._output.write_line("")

                self._output.write_line(
                    f"  <fg=red;options=bold>• "
                    f"{self._group_line(group)} "
                    f"<fg=default;options=dark>></> "
                    f"<fg=red;options=bold>{test.description}</>"
                    "</>"
                )
                self._output.write_line("    " + self._identifier_line(test.identifier))
                self._write_error(test)
                i += 1

    def _write_error(self, test: TestResult) -> None:
        self._write_parameters(test)

        if test.stdout:
            self._output.write_line("")
            self._output.write_line("    <fg=default;options=bold>Stdout:</>")
            self._output.write_line("")
            self._output.write_line(
                "    "
                + self._output.formatter.escape(test.stdout).replace("\n", "\n    ")
            )

        error = test.error
        if isinstance(error, ExceptionInfo):
            error = error.value

        trace = ExceptionTrace(error, base_indent=2)
        trace.ignore_files_in(self._ignore_files_in)
        trace.render(self._output)

        self._output.write_line("")

    def _type_color(self, type: str) -> str:
        if type == TestResult.FAIL:
            return "red"
        elif type == TestResult.SKIPPED:
            return "yellow"

        return "green"

    def _title_line(self, fg: str, bg: str, title: str, group: Tuple[str]) -> str:
        group = self._group_line(group)

        return f"\n  <fg={bg};options=bold>{title}</> <fg=default;options=dark,bold>•</> {group}"

    def _group_line(self, group: Tuple[str]) -> str:
        path = group[0]
        path_parts = path.split(os.path.sep)
        for i, part in enumerate(path_parts[:-1]):
            path_parts[i] = f"<fg=default>{part}</>"

        path_parts[-1] = f"<fg=default;options=bold>{path_parts[-1]}</>"
        line = f"<fg=default;options=dark>{os.path.sep}</>".join(path_parts)

        if len(group) > 1:
            line += " <fg=default;options=dark>></> "
            line += f"<fg=default;options=bold>{group[1]}</>"
            line += " <fg=default;options=dark>></> ".join(
                f"<fg=default;options=bold>{g}" for g in group[1:-1]
            )

        return line

    def _test_line(
        self,
        fg: str,
        icon: str,
        description: str,
        warning: Optional[str] = None,
    ) -> str:
        if warning:
            warning = (
                f" <fg=default;options=dark>(<fg=default;options=bold>Skipped</>: "
                f"<fg=yellow;options=bold>{warning}</>)</>"
            )
        else:
            warning = ""

        return (
            f"  <fg={fg};options=bold>{icon}</><fg=default> {description}{warning}</>"
        )

    def _identifier_line(self, identifier: str) -> str:
        return f"<fg=default;options=dark>{identifier}</>"

    def _footer_content(self, state: State, report: Optional[TestReport] = None) -> str:
        runs = []

        if report:
            runs.append(self._title_line("black", "blue", "RUNS", state.group))

            test = TestResult.create_from_test_report(report)
            description = test.description
            if test.parameters:
                description += " <fg=blue><b>#</b>{}</blue>".format(
                    state.parameter_set(test)
                )

            runs.append(self._test_line(test.color, test.icon, description))

        tests = []
        for type in self._TYPES:
            tests_count = state.count_tests_in_suite_with_type(type)
            if tests_count:
                color = self._type_color(type)
                tests.append(
                    f"<fg={color}><fg={color};option=bold>{tests_count}</> {type}</>"
                )

        pending = state.suite_total_tests - state.suite_tests_count
        if pending:
            tests.append(
                f"<fg=default;options=dark><fg=default;options=dark,bold>{pending}</> pending</>"
            )

        if tests:
            footer = "\n".join(
                runs
                + [""]
                + [
                    f"  <fg=default;options=bold>Tests:  </><fg=default>{', '.join(tests)}</>"
                ]
            )

            return footer

        return ""

    def _write_parameters(self, test: TestResult) -> None:
        if not test.parameters:
            return

        self._output.write_line("")
        for name, value in test.parameters:
            color = "default"
            if isinstance(value, (int, float)):
                color = "blue"
            elif isinstance(value, (str, bytes)):
                color = "yellow"

            param_description = (
                f"    <fg=default;option=bold>{name}</><fg=default;options=dark>:</> "
            )
            param_description += f"<fg={color}>{Formatter.escape(repr(value))}</>"
            self._output.write_line(param_description)

        self._output.write_line("")



import re

from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

from _pytest.reports import TestReport


class TestResult:

    FAIL = "failed"
    SKIPPED = "skipped"
    RUNS = "pending"
    PASS = "passed"

    def __init__(
        self,
        identifier: str,
        description: str,
        type: str,
        icon: str,
        color: str,
        action: str,
        error: Exception = None,
        stdout: str = None,
        skip_reason: Optional[str] = None,
        parameters: Optional[List[Tuple[str, Any]]] = None,
    ) -> None:
        self._identifier = identifier
        self._description = description
        self._type = type
        self._icon = icon
        self._color = color
        self._action = action
        self._error = error
        self._stdout = stdout
        self._skip_reason = skip_reason

        if parameters is None:
            parameters = []

        self._parameters: List[Tuple[str, Any]] = parameters

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def base_identifier(self) -> str:
        return re.sub(r"\[(.+)\]$", "", self._identifier)

    @property
    def group(self) -> str:
        return self._group

    @property
    def description(self) -> str:
        return self._description

    @property
    def type(self) -> str:
        return self._type

    @property
    def icon(self) -> str:
        return self._icon

    @property
    def color(self) -> str:
        return self._color

    @property
    def action(self) -> str:
        return self._action

    @property
    def error(self) -> Optional[Exception]:
        return self._error

    @property
    def stdout(self) -> Optional[str]:
        return self._stdout

    @property
    def skip_reason(self) -> Optional[str]:
        return self._skip_reason

    def is_skipped(self) -> bool:
        return self._type == self.SKIPPED

    @property
    def parameters(self) -> List[Tuple[str, Any]]:
        return self._parameters

    @classmethod
    def create_from_test_report(
        cls, report: TestReport, type: Optional[str] = None
    ) -> "TestResult":
        identifier = report.nodeid
        description = cls.make_description(report)
        icon = cls.make_icon(report)
        color = cls.make_color(report)
        type = type or cls.make_type(report)
        action = report.when
        stdout = None
        if hasattr(report, "sections"):
            for name, content in report.sections:
                if "stdout" in name:
                    stdout = content

        return TestResult(
            identifier,
            description,
            type,
            icon,
            color,
            action,
            error=report.exception if type == cls.FAIL else None,
            stdout=stdout,
            skip_reason=report.longrepr[2][9:]
            if report.outcome == cls.SKIPPED
            else None,
            parameters=list(report.parameters.items()),
        )

    @classmethod
    def make_description(cls, record: TestReport) -> str:
        description = record.description
        if not description:
            description = record.nodeid.split("::")[-1]

            # Replace underscores with spaces
            description = description.replace("_", " ")

            # If it starts with `test_`, we remove it
            description = re.sub("^test(.*)", "\\1", description)

            # Remove spaces
            description = description.strip()

            # Drop parameters since they will be handle by the Style instance
            description = re.sub(r"\[(.+)\]$", "", description)

        return description

    @classmethod
    def make_icon(cls, record: TestReport) -> str:
        if record.outcome == cls.FAIL:
            return "✕"
        elif record.outcome == cls.SKIPPED:
            return "s"
        elif record.when in ["setup", "call"]:
            return "•"

        return "✓"

    @classmethod
    def make_color(cls, record: TestReport) -> str:
        if record.outcome == cls.FAIL:
            return "red"
        elif record.outcome == cls.SKIPPED:
            return "yellow"
        elif record.when in ["setup", "call"]:
            return "blue"

        return "green"

    @classmethod
    def make_type(cls, record: TestReport) -> str:
        if record.outcome == cls.FAIL:
            return cls.FAIL
        elif record.outcome == cls.SKIPPED:
            return cls.SKIPPED
        elif record.when in ["setup", "call"]:
            return cls.RUNS

        return cls.PASS



from types import TracebackType
from typing import TYPE_CHECKING
from typing import Generic
from typing import Optional
from typing import Pattern
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union
from typing import cast

from _pytest.outcomes import fail


_E = TypeVar("_E", bound=BaseException)

if TYPE_CHECKING:
    from _pytest._code import ExceptionInfo


class RaisesContext(Generic[_E]):
    def __init__(
        self,
        expected_exception: Union["Type[_E]", Tuple["Type[_E]", ...]],
        message: str,
        match_expr: Optional[Union[str, "Pattern"]] = None,
    ) -> None:
        self.expected_exception = expected_exception
        self.message = message
        self.match_expr = match_expr
        self.excinfo: Optional["ExceptionInfo[_E]"] = None

    def __enter__(self) -> "ExceptionInfo[_E]":
        from _pytest._code import ExceptionInfo

        self.excinfo = ExceptionInfo.for_later()

        return self.excinfo

    def __exit__(
        self,
        exc_type: Optional[Type[_E]],
        exc_val: Optional[_E],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        __tracebackhide__ = True
        if exc_type is None:
            fail(self.message)
        assert self.excinfo is not None
        if not issubclass(exc_type, self.expected_exception):
            return False
        # Cast to narrow the exception type now that it's verified.
        exc_info = cast(
            Tuple["Type[_E]", _E, TracebackType], (exc_type, exc_val, exc_tb)
        )
        self.excinfo.fill_unfilled(exc_info)
        if self.match_expr is not None:
            self.excinfo.match(self.match_expr)
        return True



from typing import Optional
from typing import Tuple

from _pytest._code.code import ExceptionInfo
from _pytest.nodes import Item
from _pytest.outcomes import skip
from _pytest.reports import TestReport as BaseTestReport


class TestReport(BaseTestReport):
    def __init__(
        self,
        nodeid,
        location: Tuple[str, Optional[int], str],
        keywords,
        outcome,
        longrepr,
        when,
        sections=(),
        duration=0,
        user_properties=None,
        description=None,
        **extra
    ) -> None:
        super().__init__(
            nodeid,
            location,
            keywords,
            outcome,
            longrepr,
            when,
            sections=sections,
            duration=duration,
            user_properties=user_properties,
            description=description,
            **extra
        )

    @classmethod
    def from_item_and_call(cls, item: Item, call) -> "TestReport":
        callspec = item.callspec if hasattr(item, "callspec") else None
        parameters = callspec.params if callspec else {}
        when = call.when
        duration = call.stop - call.start
        keywords = {x: 1 for x in item.keywords}
        excinfo = call.excinfo
        sections = []
        exception = call.excinfo
        if not call.excinfo:
            outcome = "passed"
            longrepr = None
        else:
            if not isinstance(excinfo, ExceptionInfo):
                outcome = "failed"
                longrepr = excinfo
            elif excinfo.errisinstance(skip.Exception):
                outcome = "skipped"
                r = excinfo._getreprcrash()
                longrepr = (str(r.path), r.lineno, r.message)
            else:
                outcome = "failed"
                if call.when == "call":
                    longrepr = item.repr_failure(excinfo)
                else:  # exception in setup or teardown
                    longrepr = item._repr_failure_py(
                        excinfo, style=item.config.getoption("tbstyle", "auto")
                    )
        for rwhen, key, content in item._report_sections:
            sections.append(("Captured {} {}".format(key, rwhen), content))

        description = item.obj.__doc__
        if description is not None:
            description_parts = description.strip().split("\n", 1)
            description = description_parts[0]
            description = description.strip()

        return cls(
            item.nodeid,
            item.location,
            keywords,
            outcome,
            longrepr,
            when,
            sections,
            duration,
            user_properties=item.user_properties,
            exception=exception,
            parameters=parameters,
            description=description,
        )



import sys

from typing import TYPE_CHECKING
from typing import Any
from typing import Optional

import pytest

from _pytest._code.code import ExceptionChainRepr
from _pytest.config import Config
from _pytest.nodes import Collector
from _pytest.outcomes import Skipped
from _pytest.reports import CollectErrorRepr
from _pytest.reports import CollectReport
from _pytest.runner import CallInfo
from cleo.formatters.formatter import Formatter
from cleo.io.inputs.string_input import StringInput
from cleo.io.io import IO
from cleo.io.outputs.output import Verbosity
from cleo.io.outputs.stream_output import StreamOutput
from cleo.ui.exception_trace import ExceptionTrace

from treat import test
from treat.utils.diff import Diff

from ..ui.coverage_reporter import Coverage
from .terminal_reporter import TerminalReporter
from .test_report import TestReport


if TYPE_CHECKING:
    from _pytest._code.code import ExceptionInfo
    from _pytest._code.code import ExceptionRepr
    from _pytest.python import PyCollector


def pytest_addoption(parser):
    group = parser.getgroup("cov", "coverage reporting")

    group.addoption(
        "--cov",
        action="append",
        default=[],
        metavar="SOURCE",
        nargs="?",
        const=True,
        dest="cov_source",
        help="Path or package name to measure during execution (multi-allowed). "
        "Use --cov= to not do any source filtering and record everything.",
    )


@pytest.mark.tryfirst
def pytest_load_initial_conftests(early_config, parser, args):
    options = early_config.known_args_namespace

    io = IO(StringInput(""), StreamOutput(sys.stdout), StreamOutput(sys.stderr))
    coverage = None
    if options.cov_source:
        coverage = Coverage(io, source=early_config.known_args_namespace.cov_source)
        coverage.start()

    early_config.pluginmanager.register(TreatPlugin(io, coverage), "treatplugin")


class TreatPlugin:
    def __init__(self, io: "IO", coverage: "Coverage" = None) -> None:
        self._coverage = coverage
        self._io = io

    def pytest_pycollect_makeitem(
        self, collector: "PyCollector", name: str, obj: object
    ):
        # Avoid test() from being picked up by pytest
        if obj is test:
            return []

    def pytest_runtest_makereport(self, item, call):
        return TestReport.from_item_and_call(item, call)

    def pytest_make_collect_report(self, collector: Collector) -> CollectReport:
        call = CallInfo.from_call(lambda: list(collector.collect()), "collect")
        longrepr = None
        if not call.excinfo:
            outcome = "passed"
        else:
            skip_exceptions = [Skipped]
            unittest = sys.modules.get("unittest")
            if unittest is not None:
                # Type ignored because unittest is loaded dynamically.
                skip_exceptions.append(unittest.SkipTest)  # type: ignore
            if call.excinfo.errisinstance(tuple(skip_exceptions)):
                outcome = "skipped"
                r_ = collector._repr_failure_py(call.excinfo, "line")
                assert isinstance(r_, ExceptionChainRepr), repr(r_)
                r = r_.reprcrash
                assert r
                longrepr = (str(r.path), r.lineno, r.message)
            else:
                outcome = "failed"
                errorinfo = collector.repr_failure(call.excinfo)
                if not hasattr(errorinfo, "toterminal"):
                    errorinfo = CollectErrorRepr(errorinfo)
                longrepr = errorinfo

        rep = CollectReport(
            collector.nodeid,
            outcome,
            longrepr,
            getattr(call, "result", None),
            excinfo=call.excinfo,
        )
        rep.call = call  # type: ignore # see collect_one_node

        return rep

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtestloop(self, session):
        yield

        if self._coverage:
            self._coverage.stop()

        reporter = session.config.pluginmanager.getplugin("terminalreporter")
        if reporter:
            reporter._count = len(session.items)

    @pytest.mark.trylast
    def pytest_configure(self, config):
        current_reporter = config.pluginmanager.getplugin("terminalreporter")
        config.pluginmanager.unregister(current_reporter)
        terminal_reporter = TerminalReporter(
            current_reporter.config, coverage=self._coverage
        )
        config.pluginmanager.register(terminal_reporter, "terminalreporter")

    def pytest_internalerror(
        self,
        excrepr: "ExceptionRepr",
        excinfo: "ExceptionInfo[BaseException]",
    ) -> Optional[bool]:
        io = StreamOutput(sys.stdout)
        io.set_verbosity(Verbosity.VERBOSE)
        trace = ExceptionTrace(excinfo.value)

        trace.render(io)

    # Better diffs display
    def pytest_assertrepr_compare(
        self, config: Config, op: "str", left: Any, right: Any
    ):
        if op != "==":
            return

        diff = Diff()
        output = diff.terminal_diff(left, right)

        formatter = Formatter(config.get_terminal_writer().hasmarkup)

        return [
            formatter.format(
                "<fg=red;options=bold>{}</> <fg=default;options=dark>==</> <fg=green;options=bold>{}</>".format(
                    formatter.escape(repr(left)), formatter.escape(repr(right))
                )
            ),
            "",
        ] + [formatter.format(o) for o in output]



from collections import defaultdict
from typing import Dict
from typing import Iterator
from typing import Tuple

from .test_result import TestResult


class State:
    def __init__(self, group: Tuple[str]) -> None:
        self.suite_total_tests = None
        self._suite_tests = {}
        self._group = group
        self._group_tests: Dict[str, TestResult] = {}
        self._parameter_sets = defaultdict(dict)

    @property
    def group(self) -> Tuple[str]:
        return self._group

    @property
    def tests(self) -> Iterator[TestResult]:
        return self._group_tests.values()

    @property
    def group_title(self) -> str:
        if any(test.type == TestResult.FAIL for test in self._group_tests.values()):
            return "FAIL"

        if any(test.type != TestResult.PASS for test in self._group_tests.values()):
            return "WARN"

        return "PASS"

    @property
    def group_title_color(self) -> str:
        if any(test.type == TestResult.FAIL for test in self._group_tests.values()):
            return "red"

        if any(test.type != TestResult.PASS for test in self._group_tests.values()):
            return "yellow"

        return "green"

    @property
    def group_tests_count(self) -> int:
        return len(self._group_tests)

    @property
    def suite_tests_count(self) -> int:
        count = 0
        for group_tests in self._suite_tests.values():
            for _ in group_tests.values():
                count += 1

        return count

    def parameter_set(self, test) -> int:
        return self._parameter_sets[test.base_identifier][test.identifier]

    def add(self, test: TestResult) -> None:
        if test.identifier not in self._group_tests:
            self._group_tests[test.identifier] = test
        else:
            current_test = self._group_tests[test.identifier]
            if current_test.type != test.type:
                if test.type == TestResult.FAIL:
                    self._group_tests[test.identifier] = test

        if self._group not in self._suite_tests:
            self._suite_tests[self._group] = {}

        self._suite_tests[self._group] = self._group_tests

    def test_exists_in_group(self, test: TestResult) -> bool:
        return test.identifier in self._group_tests

    def group_has_changed(self, test: TestResult) -> bool:
        return tuple(test.identifier.split("::")[:-1]) != self._group

    def move_to(self, test: TestResult) -> None:
        self._group = tuple(test.identifier.split("::")[:-1])
        self._group_tests = {}

    def count_tests_in_suite_with_type(self, type: str) -> int:
        count = 0

        for group_tests in self._suite_tests.values():
            for test in group_tests.values():
                if test.type == type:
                    count += 1

        return count

    def add_parameter_set_for(self, test: TestResult) -> None:
        if test.parameters:
            if test.identifier not in self._parameter_sets[test.base_identifier]:
                self._parameter_sets[test.base_identifier][test.identifier] = (
                    len(self._parameter_sets[test.base_identifier]) + 1
                )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['treat', 'treat.mock', 'treat.pytest', 'treat.ui', 'treat.utils']

package_data = \
{'': ['*']}

install_requires = \
['cleo>=1.0.0a3,<2.0.0',
 'coverage>=6.2,<7.0',
 'pprintpp>=0.4.0,<0.5.0',
 'pytest>=6.2.5,<7.0.0']

entry_points = \
{'pytest11': ['treat = treat.pytest.plugin']}

setup_kwargs = {
    'name': 'treat',
    'version': '0.1.0',
    'description': 'Delicious tests for Python',
    'long_description': 'None',
    'author': 'Sébastien Eustace',
    'author_email': 'sebastien@eustace.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
